@@ -17,7 +17,7 @@ not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 Cambridge, MA 02139, USA.  */
 
 /*
- *	POSIX Standard: 2.10 Symbolic Constants		<unistd.h>
+ *	POSIX Standard: 2.9 Symbolic Constants		<unistd.h>
  */
 
 #ifndef	_UNISTD_H
@@ -30,8 +30,8 @@ __BEGIN_DECLS
 /* These may be used to determine what facilities are present at compile time.
    Their values can be obtained at run time from sysconf.  */
 
-/* POSIX Standard approved as IEEE Std 1003.1 as of August, 1988.  */
-#define	_POSIX_VERSION	199009L
+/* POSIX Standard approved as IEEE Std 1003.1b as of September 15, 1993.  */
+#define	_POSIX_VERSION	199309L
 
 /* These are not #ifdef __USE_POSIX2 because they are
    in the theoretically application-owned namespace.  */
@@ -439,12 +439,16 @@ extern __pid_t vfork __P ((void));
 
 #endif /* Use BSD. */
 
+#if 0
+
 /* get character login name of the user. */
 extern char *cuserid __P((char * __buf));
 
 /* Return the name of the controlling terminal. */
 extern char *ctermid __P((char *__buf));
 
+#endif
+
 /* Return the pathname of the terminal FD is open on, or NULL on errors.
    The returned storage is good only until the next call to this function.  */
 extern char *ttyname __P ((int __fd));
@@ -562,6 +566,9 @@ extern size_t getpagesize __P ((void));
 extern int __getdtablesize __P ((void));
 extern int getdtablesize __P ((void));
 
+/* Make all changes done to FD actually appear on disk, except changes
+ * to the access times in the inode.  */
+extern int fdatasync __P ((int __fd));
 
 /* Make all changes done to FD actually appear on disk.  */
 extern int fsync __P ((int __fd));
@@ -587,7 +594,7 @@ extern int chroot __P ((__const char *__path));
 
 /* Make the block special device PATH available to the system for swapping.
    This call is restricted to the super-user.  */
-extern int swapon __P ((__const char *__path));
+extern int swapon __P ((__const char *__path, int __swapflags));
 
 /* Successive calls return the shells listed in `/etc/shells'.  */
 extern char *getusershell __P ((void));
