@@ -1,4 +1,4 @@
-/* Copyright (C) 1991, 1992, 1993, 1994, 1995 Free Software Foundation, Inc.
+/* Copyright (C) 1991, 92, 93, 94, 95, 96 Free Software Foundation, Inc.
 This file is part of the GNU C Library.
 
 The GNU C Library is free software; you can redistribute it and/or
@@ -34,29 +34,33 @@ Cambridge, MA 02139, USA.  */
 #define	LONGLONG	long
 #endif
 
+/* Those are flags in the conversion format. */
+# define LONG		0x001	/* l: long or double */
+# define LONGDBL	0x002	/* L: long long or long double */
+# define SHORT		0x004	/* h: short */
+# define SUPPRESS	0x008	/* *: suppress assignment */
+# define POINTER	0x010	/* weird %p pointer (`fake hex') */
+# define NOSKIP		0x020	/* do not skip blanks */
+# define WIDTH		0x040	/* width was given */
+# define GROUP		0x080	/* ': group numbers */
+# define MALLOC		0x100	/* a: malloc strings */
+
+# define TYPEMOD	(LONG|LONGDBL|SHORT)
+
+
 #ifdef USE_IN_LIBIO
 # include <libioP.h>
 # include <libio.h>
 
-/* Those are flags in the conversion format. */
-# define LONG		0x01	/* l: long or double */
-# define LONGDBL	0x02	/* L: long long or long double */
-# define SHORT		0x04	/* h: short */
-# define SUPPRESS	0x08	/* suppress assignment */
-# define POINTER	0x10	/* weird %p pointer (`fake hex') */
-# define NOSKIP		0x20	/* do not skip blanks */
-# define WIDTH		0x40	/* width */
-
-
 # define va_list	_IO_va_list
 # define ungetc(c, s)	_IO_ungetc (c, s)
-# define inchar()	((c = _IO_getc(s)), ++read_in, c)
-# define conv_error()	return ((errp != NULL && (*errp |= 2)), \
-				(c == EOF || _IO_ungetc(c, s)), done)
+# define inchar()	((c = _IO_getc (s)), (void) ++read_in, c)
+# define conv_error()	return ((void) (errp != NULL && (*errp |= 2)), \
+				(void) (c == EOF || _IO_ungetc (c, s)), done)
 
-# define input_error()	return ((errp != NULL && (*errp |= 1)), \
+# define input_error()	return ((void) (errp != NULL && (*errp |= 1)), \
 				done == 0 ? EOF : done)
-# define memory_error()	return ((errno = ENOMEM), EOF)
+# define memory_error()	return ((void) (errno = ENOMEM), EOF)
 # define ARGCHECK(s, format)						     \
   do									     \
     {									     \
@@ -69,10 +73,10 @@ Cambridge, MA 02139, USA.  */
        }								     \
     } while (0)
 #else
-# define inchar()	((c = getc(s)) == EOF ? EOF : (++read_in, c))
-# define conv_error()	return (ungetc(c, s), done)
+# define inchar()	((c = getc (s)), (void) ++read_in, c)
+# define conv_error()	return ((void) ungetc (c, s), done)
 # define input_error()	return (done == 0 ? EOF : done)
-# define memory_error()	return ((errno = ENOMEM), EOF)
+# define memory_error()	return ((void) (errno = ENOMEM), EOF)
 # define ARGCHECK(s, format)						     \
   do									     \
     {									     \
@@ -98,43 +102,31 @@ _IO_vfscanf (s, format, argptr, errp)
      int *errp;
 #else
 int
-__vfscanf (s, format, arg)
-     FILE *s;
-     const char *format;
-     va_list argptr;
+__vfscanf (FILE *s, const char *format, va_list argptr)
 #endif
 {
   va_list arg = (va_list) argptr;
 
   register const char *f = format;
-  register char fc;		/* Current character of the format.  */
+  register unsigned char fc;	/* Current character of the format.  */
   register size_t done = 0;	/* Assignments done.  */
   register size_t read_in = 0;	/* Chars read in.  */
   register int c;		/* Last char read.  */
-  register int do_assign;	/* Whether to do an assignment.  */
   register int width;		/* Maximum field width.  */
-  int group_flag;		/* %' modifier flag.  */
-#ifdef USE_IN_LIBIO
-  int flags;			/* Trace flags for current format element.  */
-#endif
+  register int flags;		/* Modifiers for current format element.  */
 
-  /* Type modifiers.  */
-  int is_short, is_long, is_long_double;
-#ifdef	HAVE_LONGLONG
-  /* We use the `L' modifier for `long long int'.  */
-# define is_longlong	is_long_double
-#else
-# define is_longlong	0
-#endif
-  int malloc_string;		/* Args are char ** to be filled in.  */
   /* Status for reading F-P nums.  */
-  char got_dot, got_e;
+  char got_dot, got_e, negative;
   /* If a [...] is a [^...].  */
   char not_in;
   /* Base for integral numbers.  */
   int base;
   /* Signedness for integral numbers.  */
   int number_signed;
+  /* Decimal point character.  */
+  wchar_t decimal;
+  /* The thousands character of the current locale.  */
+  wchar_t thousands;
   /* Integral holding variables.  */
   union
     {
@@ -146,10 +138,29 @@ __vfscanf (s, format, arg)
   /* Character-buffer pointer.  */
   register char *str, **strptr;
   size_t strsize;
+  /* We must not react on white spaces immediately because they can
+     possibly be matched even if in the input stream no character is
+     available anymore.  */
+  int skip_space = 0;
   /* Workspace.  */
-  char work[200];
-  char *w;			/* Pointer into WORK.  */
-  wchar_t decimal;		/* Decimal point character.  */
+  char *tw;			/* Temporary pointer.  */
+  char *wp = NULL;		/* Workspace.  */
+  size_t wpmax = 0;		/* Maximal size of workspace.  */
+  size_t wpsize;		/* Currently used bytes in workspace.  */
+#define ADDW(Ch)							    \
+  do									    \
+    {									    \
+      if (wpsize == wpmax)						    \
+	{								    \
+	  char *old = wp;						    \
+	  wpmax = UCHAR_MAX > 2 * wpmax ? UCHAR_MAX : 2 * wpmax;	    \
+	  wp = (char *) alloca (wpmax);					    \
+	  if (old != NULL)						    \
+	    memcpy (wp, old, wpsize);					    \
+	}								    \
+      wp[wpsize++] = (Ch);						    \
+    }									    \
+  while (0)
 
   ARGCHECK (s, format);
 
@@ -157,6 +168,10 @@ __vfscanf (s, format, arg)
   if (mbtowc (&decimal, _NL_CURRENT (LC_NUMERIC, DECIMAL_POINT),
 	      strlen (_NL_CURRENT (LC_NUMERIC, DECIMAL_POINT))) <= 0)
     decimal = (wchar_t) *_NL_CURRENT (LC_NUMERIC, DECIMAL_POINT);
+  /* Figure out the thousands separator character.  */
+  if (mbtowc (&thousands, _NL_CURRENT (LC_NUMERIC, THOUSANDS_SEP),
+	      strlen (_NL_CURRENT (LC_NUMERIC, THOUSANDS_SEP))) <= 0)
+    thousands = (wchar_t) *_NL_CURRENT (LC_NUMERIC, THOUSANDS_SEP);
 
   c = inchar ();
 
@@ -206,42 +221,42 @@ __vfscanf (s, format, arg)
       fc = *f++;
       if (fc != '%')
 	{
-	  /* Characters other than format specs must just match.  */
-#if 0
-	  if (c == EOF)
-	    input_error ();
-#endif
-	  /* We may get a special case here. If %n is the next
-	   * one. It doesn't matter if there is no more char in
-	   * the input string. H.J. */
+	  /* Remember to skip spaces.  */
 	  if (isspace (fc))
 	    {
-	      /* Whitespace characters match any amount of whitespace.  */
-	      while (isspace (c))
-		inchar ();
+	      skip_space = 1;
 	      continue;
 	    }
-	  else if (c == fc)
-	    (void) inchar ();
-	  /* We need to check EOF here but only if we have nothing to
-	     convert. "%n" can have zero input. H.J. */
-	  else if (c == EOF && read_in == 1)
+
+	  /* Characters other than format specs must just match.  */
+	  if (c == EOF)
 	    input_error ();
+
+	  /* We saw white space char as the last character in the format
+	     string.  Now it's time to skip all leading white space.  */
+	  if (skip_space)
+	    {
+	      while (isspace (c))
+		(void) inchar ();
+	      skip_space = 0;
+	    }
+
+	  if (c == fc)
+	    (void) inchar ();
 	  else
 	    conv_error ();
+
 	  continue;
 	}
 
-#ifdef USE_IN_LIBIO
-      /* That is the start of the coversion string. */
+      /* This is the start of the conversion string. */
       flags = 0;
-#endif
 
       /* Initialize state of modifiers.  */
       argpos = 0;
-      do_assign = 1;
-      group_flag = 0;
-      is_short = is_long = is_long_double = malloc_string = 0;
+
+      /* Prepare temporary buffer.  */
+      wpsize = 0;
 
       /* Check for a positional parameter specification.  */
       if (isdigit (*f))
@@ -255,6 +270,7 @@ __vfscanf (s, format, arg)
 	    {
 	      /* Oops; that was actually the field width.  */
 	      width = argpos;
+	      flags |= WIDTH;
 	      argpos = 0;
 	      goto got_width;
 	    }
@@ -265,21 +281,16 @@ __vfscanf (s, format, arg)
 	switch (*f++)
 	  {
 	  case '*':
-#ifdef USE_IN_LIBIO
-	    flags = SUPPRESS;
-#endif
-	    do_assign = 0;
+	    flags |= SUPPRESS;
 	    break;
 	  case '\'':
-	    group_flag = 1;
+	    flags |= GROUP;
 	    break;
 	  }
 
-#ifdef USE_IN_LIBIO
       /* We have seen width. */
       if (isdigit (*f))
 	flags |= WIDTH;
-#endif
 
       /* Find the maximum field width.  */
       width = 0;
@@ -298,50 +309,39 @@ __vfscanf (s, format, arg)
 	  {
 	  case 'h':
 	    /* int's are short int's.  */
-#ifdef USE_IN_LIBIO
-	    if (flags & ~(SUPPRESS | WIDTH))
+	    if (flags & TYPEMOD)
 	      /* Signal illegal format element.  */
 	      conv_error ();
 	    flags |= SHORT;
-#endif
-	    is_short = 1;
 	    break;
 	  case 'l':
-	    if (is_long)
+	    if (flags & (SHORT|LONGDBL))
+	      conv_error ();
+	    else if (flags & LONG)
 	      {
 		/* A double `l' is equivalent to an `L'.  */
-#ifdef USE_IN_LIBIO
-		if ((flags & ~(SUPPRESS | WIDTH)) && (flags & LONGDBL))
-		  conv_error ();
 		flags &= ~LONG;
 		flags |= LONGDBL;
-#endif
-		is_longlong = 1;
 	      }
 	    else
-	      {
-		/* int's are long int's.  */
-#ifdef USE_IN_LIBIO
-		flags |= LONG;
-#endif
-		is_long = 1;
-	      }
+	      /* int's are long int's.  */
+	      flags |= LONG;
 	    break;
 	  case 'q':
 	  case 'L':
 	    /* double's are long double's, and int's are long long int's.  */
-#ifdef USE_IN_LIBIO
-	    if (flags & ~(SUPPRESS | WIDTH))
+	    if (flags & TYPEMOD)
 	      /* Signal illegal format element.  */
 	      conv_error ();
 	    flags |= LONGDBL;
-#endif
-	    is_long_double = 1;
 	    break;
 	  case 'a':
+	    if (flags & TYPEMOD)
+	      /* Signal illegal format element.  */
+	      conv_error ();
 	    /* String conversions (%s, %[) take a `char **'
 	       arg and fill it in with a malloc'd pointer.  */
-	    malloc_string = 1;
+	    flags |= MALLOC;
 	    break;
 	  }
 
@@ -350,25 +350,15 @@ __vfscanf (s, format, arg)
 	conv_error ();
 
       /* Find the conversion specifier.  */
-      w = work;
       fc = *f++;
+      if (skip_space || (fc != '[' && fc != 'c' && fc != 'n'))
+	{
+	  /* Eat whitespace.  */
+	  while (isspace (c))
+	    (void) inchar ();
+	  skip_space = 0;
+	}
 
-      if (fc == 'n')
-      {
-	/* Answer number of assignments done.  */
-	if (do_assign)
-	  *ARG (int *) = read_in - 1;	/* Don't count the read-ahead.  */
-	continue;
-      }
-
-      /* End of the input string?  */
-      if (c == EOF)
-	input_error ();
-
-      if (fc != '[' && fc != 'c')
-	/* Eat whitespace.  */
-	while (isspace (c))
-	  (void) inchar ();
       switch (fc)
 	{
 	case '%':	/* Must match a literal '%'.  */
@@ -377,15 +367,23 @@ __vfscanf (s, format, arg)
 	  inchar ();
 	  break;
 
-#if 0
 	case 'n':	/* Answer number of assignments done.  */
-	  if (do_assign)
-	    *ARG (int *) = read_in - 1;	/* Don't count the read-ahead.  */
+	  /* Corrigendum 1 to ISO C 1990 describes the allowed flags
+	     with the 'n' conversion specifier.  */
+	  if (!(flags & SUPPRESS))
+	    /* Don't count the read-ahead.  */
+	    if (flags & LONGDBL)
+	      *ARG (long long int *) = read_in - 1;
+	    else if (flags & LONG)
+	      *ARG (long int *) = read_in - 1;
+	    else if (flags & SHORT)
+	      *ARG (short int *) = read_in - 1;
+	    else
+	      *ARG (int *) = read_in - 1;
 	  break;
-#endif
 
 	case 'c':	/* Match characters.  */
-	  if (do_assign)
+	  if (!(flags & SUPPRESS))
 	    {
 	      str = ARG (char *);
 	      if (str == NULL)
@@ -398,7 +396,7 @@ __vfscanf (s, format, arg)
 	  if (width == -1)
 	    width = 1;
 
-	  if (do_assign)
+	  if (!(flags & SUPPRESS))
 	    {
 	      do
 		*str++ = c;
@@ -407,16 +405,16 @@ __vfscanf (s, format, arg)
 	  else
 	    while (inchar () != EOF && --width > 0);
 
-	  if (do_assign)
+	  if (!(flags & SUPPRESS))
 	    ++done;
 
 	  break;
 
 	case 's':		/* Read a string.  */
 #define STRING_ARG							      \
-	  if (do_assign)						      \
+	  if (!(flags & SUPPRESS))					      \
 	    {								      \
-	      if (malloc_string)					      \
+	      if (flags & MALLOC)					      \
 		{							      \
 		  /* The string is to be stored in a malloc'd buffer.  */     \
 		  strptr = ARG (char **);				      \
@@ -441,10 +439,10 @@ __vfscanf (s, format, arg)
 	      if (isspace (c))
 		break;
 #define	STRING_ADD_CHAR(c)						      \
-	      if (do_assign)						      \
+	      if (!(flags & SUPPRESS))					      \
 		{							      \
 		  *str++ = c;						      \
-		  if (malloc_string && str == *strptr + strsize)	      \
+		  if ((flags & MALLOC) && str == *strptr + strsize)	      \
 		    {							      \
 		      /* Enlarge the buffer.  */			      \
 		      str = realloc (*strptr, strsize * 2);		      \
@@ -456,7 +454,7 @@ __vfscanf (s, format, arg)
 			    {						      \
 			      /* We lose.  Oh well.			      \
 				 Terminate the string and stop converting,    \
-				 so at least we don't swallow any input.  */  \
+				 so at least we don't skip any input.  */     \
 			      (*strptr)[strsize] = '\0';		      \
 			      ++done;					      \
 			      conv_error ();				      \
@@ -479,7 +477,7 @@ __vfscanf (s, format, arg)
 	      STRING_ADD_CHAR (c);
 	    } while (inchar () != EOF && (width <= 0 || --width > 0));
 
-	  if (do_assign)
+	  if (!(flags & SUPPRESS))
 	    {
 	      *str = '\0';
 	      ++done;
@@ -518,26 +516,22 @@ __vfscanf (s, format, arg)
 	  /* Check for a sign.  */
 	  if (c == '-' || c == '+')
 	    {
-	      *w++ = c;
+	      ADDW (c);
 	      if (width > 0)
 		--width;
 	      (void) inchar ();
 	    }
 
 	  /* Look for a leading indication of base.  */
-	  /* We have to check if width != 0. H.J. */
 	  if (width != 0 && c == '0')
 	    {
 	      if (width > 0)
 		--width;
-	      *w++ = '0';
+	      ADDW ('0');
 
 	      (void) inchar ();
 
-	      /* We have to check if width != 0. H.J. */
-	      if (width != 0)
-	      {
-		if (tolower (c) == 'x')
+	      if (width != 0 && tolower (c) == 'x')
 		{
 		  if (base == 0)
 		    base = 16;
@@ -545,66 +539,82 @@ __vfscanf (s, format, arg)
 		    {
 		      if (width > 0)
 			--width;
+		      ADDW (c);
 		      (void) inchar ();
 		    }
 		}
-		else if (base == 0)
-		  base = 8;
-	      }
+	      else if (base == 0)
+		base = 8;
 	    }
 
 	  if (base == 0)
 	    base = 10;
 
-	  /* Read the number into WORK.  */
-	  /* We have to check if width != 0. H.J. */
-	  if (width != 0)
-	  {
-	    do
+	  /* Read the number into workspace.  */
+	  while (c != EOF && width != 0)
 	    {
 	      if (base == 16 ? !isxdigit (c) :
-		  (!isdigit (c) || c - '0' >= base))
+		  ((!isdigit (c) || c - '0' >= base) &&
+		   !((flags & GROUP) && base == 10 && c == thousands)))
 		break;
-	      *w++ = c;
+	      ADDW (c);
 	      if (width > 0)
 		--width;
-	    }
-	    while (inchar () != EOF && width != 0);
-	  }
 
-	  if (w == work ||
-	      (w - work == 1 && (work[0] == '+' || work[0] == '-')))
+	      (void) inchar ();
+	    }
+
+	  if (wpsize == 0 ||
+	      (wpsize == 1 && (wp[0] == '+' || wp[0] == '-')))
 	    /* There was no number.  */
 	    conv_error ();
 
 	  /* Convert the number.  */
-	  *w = '\0';
-	  if (is_longlong)
+	  ADDW ('\0');
+	  if (flags & LONGDBL)
 	    {
 	      if (number_signed)
-		num.q = __strtoq_internal (work, &w, base, group_flag);
+		num.q = __strtoq_internal (wp, &tw, base, flags & GROUP);
 	      else
-		num.uq = __strtouq_internal (work, &w, base, group_flag);
+		num.uq = __strtouq_internal (wp, &tw, base, flags & GROUP);
 	    }
 	  else
 	    {
 	      if (number_signed)
-		num.l = __strtol_internal (work, &w, base, group_flag);
+		num.l = __strtol_internal (wp, &tw, base, flags & GROUP);
 	      else
-		num.ul = __strtoul_internal (work, &w, base, group_flag);
+		num.ul = __strtoul_internal (wp, &tw, base, flags & GROUP);
 	    }
-	  if (w == work)
+	  if (wp == tw)
 	    conv_error ();
 
-	  if (do_assign)
+#if USE_IN_LIBIO
+	  /* Perhaps we have too many characters in the workspace.
+	     This can happen in rare cases like "%i" with input "0xyz"
+	     or with wrong thousands separator placement.  */
+#define UNDO_TOO_MANY	if (*tw != '\0')			\
+			{					\
+			  /* Undo the ADDW ('\0').  */		\
+			  --wpsize;				\
+			  while (wpsize > tw - wp)		\
+			  {					\
+			    ungetc (c, s);			\
+			    c = wp[--wpsize];			\
+			  }					\
+			}
+
+	  UNDO_TOO_MANY
+#endif
+
+	  if (!(flags & SUPPRESS))
 	    {
 	      if (! number_signed)
 		{
-		  if (is_longlong)
+		  if (flags & LONGDBL)
 		    *ARG (unsigned LONGLONG int *) = num.uq;
-		  else if (is_long)
+		  else if (flags & LONG)
 		    *ARG (unsigned long int *) = num.ul;
-		  else if (is_short)
+		  else if (flags & SHORT)
 		    *ARG (unsigned short int *)
 		      = (unsigned short int) num.ul;
 		  else
@@ -612,11 +622,11 @@ __vfscanf (s, format, arg)
 		}
 	      else
 		{
-		  if (is_longlong)
+		  if (flags & LONGDBL)
 		    *ARG (LONGLONG int *) = num.q;
-		  else if (is_long)
+		  else if (flags & LONG)
 		    *ARG (long int *) = num.l;
-		  else if (is_short)
+		  else if (flags & SHORT)
 		    *ARG (short int *) = (short int) num.l;
 		  else
 		    *ARG (int *) = (int) num.l;
@@ -636,67 +646,75 @@ __vfscanf (s, format, arg)
 	  /* Check for a sign.  */
 	  if (c == '-' || c == '+')
 	    {
-	      *w++ = c;
+	      negative = c == '-';
 	      if (inchar () == EOF)
 		/* EOF is only an input error before we read any chars.  */
 		conv_error ();
 	      if (width > 0)
 		--width;
 	    }
+	  else
+	    negative = 0;
 
 	  got_dot = got_e = 0;
 	  do
 	    {
 	      if (isdigit (c))
-		*w++ = c;
-	      else if (got_e && w[-1] == 'e' && (c == '-' || c == '+'))
-		*w++ = c;
-	      else if (!got_e && tolower (c) == 'e')
+		ADDW (c);
+	      else if (got_e && wp[wpsize - 1] == 'e'
+		       && (c == '-' || c == '+'))
+		ADDW (c);
+	      else if (wpsize > 0 && !got_e && tolower (c) == 'e')
 		{
-		  *w++ = 'e';
+		  ADDW (c);
 		  got_e = got_dot = 1;
 		}
 	      else if (c == decimal && !got_dot)
 		{
-		  *w++ = c;
+		  ADDW (c);
 		  got_dot = 1;
 		}
+	      else if ((flags & GROUP) && c == thousands && !got_dot)
+		ADDW (c);
 	      else
 		break;
 	      if (width > 0)
 		--width;
-	    } while (inchar () != EOF && width != 0);
+	    }
+	  while (inchar () != EOF && width != 0);
 
-	  if (w == work)
-	    conv_error();
-	  if (w[-1] == '-' || w[-1] == '+' || w[-1] == 'e')
+	  if (wpsize == 0)
 	    conv_error ();
 
 	  /* Convert the number.  */
-	  *w = '\0';
-	  if (is_long_double)
+	  ADDW ('\0');
+	  if (flags & LONGDBL)
 	    {
-	      long double d = __strtold_internal (work, &w, group_flag);
-	      if (do_assign && w != work)
-		*ARG (long double *) = d;
+	      long double d = __strtold_internal (wp, &tw, flags & GROUP);
+	      if (!(flags & SUPPRESS) && tw != wp)
+		*ARG (long double *) = negative ? -d : d;
 	    }
-	  else if (is_long)
+	  else if (flags & LONG)
 	    {
-	      double d = __strtod_internal (work, &w, group_flag);
-	      if (do_assign && w != work)
-		*ARG (double *) = d;
+	      double d = __strtod_internal (wp, &tw, flags & GROUP);
+	      if (!(flags & SUPPRESS) && tw != wp)
+		*ARG (double *) = negative ? -d : d;
 	    }
 	  else
 	    {
-	      float d = __strtof_internal (work, &w, group_flag);
-	      if (do_assign && w != work)
-		*ARG (float *) = d;
+	      float d = __strtof_internal (wp, &tw, flags & GROUP);
+	      if (!(flags & SUPPRESS) && tw != wp)
+		*ARG (float *) = negative ? -d : d;
 	    }
 
-	  if (w == work)
+	  if (tw == wp)
 	    conv_error ();
 
-	  if (do_assign)
+#if USE_IN_LIBIO
+	  UNDO_TOO_MANY
+#endif
+
+	  if (!(flags & SUPPRESS))
 	    ++done;
 	  break;
 
@@ -714,35 +732,56 @@ __vfscanf (s, format, arg)
 	  else
 	    not_in = 0;
 
+	  /* Fill WP with byte flags indexed by character.
+	     We will use this flag map for matching input characters.  */
+	  if (wpmax < UCHAR_MAX)
+	    {
+	      wpmax = UCHAR_MAX;
+	      wp = (char *) alloca (wpmax);
+	    }
+	  memset (wp, 0, UCHAR_MAX);
+
+	  fc = *f;
+	  if (fc == ']' || fc == '-')
+	    {
+	      /* If ] or - appears before any char in the set, it is not
+		 the terminator or separator, but the first char in the
+		 set.  */
+	      wp[fc] = 1;
+	      ++f;
+	    }
+
 	  while ((fc = *f++) != '\0' && fc != ']')
 	    {
 	      if (fc == '-' && *f != '\0' && *f != ']' &&
-		  w > work && w[-1] <= *f)
-		/* Add all characters from the one before the '-'
-		   up to (but not including) the next format char.  */
-		for (fc = w[-1] + 1; fc < *f; ++fc)
-		  *w++ = fc;
+		  (unsigned char) f[-2] <= (unsigned char) *f)
+		{
+		  /* Add all characters from the one before the '-'
+		     up to (but not including) the next format char.  */
+		  for (fc = f[-2]; fc < *f; ++fc)
+		    wp[fc] = 1;
+		}
 	      else
-		/* Add the character to the list.  */
-		*w++ = fc;
+		/* Add the character to the flag map.  */
+		wp[fc] = 1;
 	    }
 	  if (fc == '\0')
 	    conv_error();
 
-	  *w = '\0';
 	  num.ul = read_in;
 	  do
 	    {
-	      if ((strchr (work, c) == NULL) != not_in)
+	      if (wp[c] == not_in)
 		break;
 	      STRING_ADD_CHAR (c);
 	      if (width > 0)
 		--width;
-	    } while (inchar () != EOF && width != 0);
+	    }
+	  while (inchar () != EOF && width != 0);
 	  if (read_in == num.ul)
 	    conv_error ();
 
-	  if (do_assign)
+	  if (!(flags & SUPPRESS))
 	    {
 	      *str = '\0';
 	      ++done;
@@ -752,15 +791,34 @@ __vfscanf (s, format, arg)
 	case 'p':	/* Generic pointer.  */
 	  base = 16;
 	  /* A PTR must be the same size as a `long int'.  */
-	  is_long = 1;
+	  flags &= ~(SHORT|LONGDBL);
+	  flags |= LONG;
 	  number_signed = 0;
 	  goto number;
 	}
     }
 
-  return ((c == EOF || ungetc (c, s)), done);
+  /* The last thing we saw int the format string was a white space.
+     Consume the last white spaces.  */
+  if (skip_space)
+    while (isspace (c))
+      (void) inchar ();
+
+  return ((void) (c == EOF || ungetc (c, s)), done);
 }
 
-#ifndef USE_IN_LIBIO
+#ifdef USE_IN_LIBIO
+int __vfscanf( FILE *, const char *, va_list );
+
+int
+__vfscanf (FILE *s, const char *format, va_list argptr)
+{
+  return _IO_vfscanf (s, format, argptr, NULL);
+}
+#endif
+
+#ifdef __ELF__
+#pragma weak vfscanf = __vfscanf
+#else
 weak_alias (__vfscanf, vfscanf)
 #endif
