@@ -3,9 +3,10 @@
 
 #undef getc
 
+int getc( FILE * ) __attribute__ ((weak));
+
 int
-getc(stream)
-  FILE *stream;
+getc( FILE *stream )
 {
 #ifdef _LIBPTHREAD
   int ret;
@@ -18,8 +19,10 @@ getc(stream)
 #endif
 }
 
+#if 0
 #ifndef _LIBPTHREAD
 #ifdef __ELF__
 #pragma weak getc
 #endif
 #endif
+#endif	/* #if 0 */
