@@ -1,6 +1,5 @@
 #include "libioP.h"
 #include "stdio.h"
-#include <gnu-stabs.h>
 
 #undef putc
 
@@ -27,5 +26,3 @@ int putc( int, FILE * ) __attribute__ ((weak));
 #endif   /* __GNUC__ */
 #endif   /* __ELF__ */
 #endif   /* _LIBPTHREAD */
-
-elf_alias (putc, _IO_putc);
