@@ -49,6 +49,12 @@ static char sccsid[] = "@(#)syslog.c	8.4 (Berkeley) 3/18/94";
  *
  * Author: Eric Allman
  * Modified to use UNIX domain IPC by Ralph Campbell
+ * Patched March 12, 1996 by A. Ian Vogelesang <vogelesang@hdshq.com>
+ *  - to correct the handling of message & format string truncation,
+ *  - to visibly tag truncated records to facilitate
+ *    investigation of such Bad Things with grep, and,
+ *  - to correct the handling of case where "write"
+ *    returns after writing only part of the message.
  */
 
 #include <sys/types.h>
@@ -71,11 +77,14 @@ static char sccsid[] = "@(#)syslog.c	8.4 (Berkeley) 3/18/94";
 #include <stdarg.h>
 #include <paths.h>
 #include <stdio.h>
+#include <ctype.h>
 
 
 #if defined(_REENTRENT) || defined(_THREAD_SAFE)
 # include <pthread.h>
 
+extern int __writev( int, const struct iovec *, size_t);
+
 /* We need to initialize the mutex.  For this we use a method provided
    by pthread function 'pthread_once'.  For this we need a once block.  */
 static pthread_once__t _once_block = pthread_once_init;
@@ -110,6 +119,18 @@ static const char *LogTag = "syslog";	/* string to tag the entry with */
 static int	LogFacility = LOG_USER;	/* default facility code */
 static int	LogMask = 0xff;		/* mask of priorities to be logged */
 
+/* #define TRUNCATED "[truncated by syslog.c] "
+*/
+#define TRUNCATED "[truncated] " /* AIV */
+
+static void closelog_intern( int );
+void syslog( int, const char *, ...);
+void vsyslog( int, const char *, va_list );
+void openlog( const char *, int, int );
+void closelog( void );
+int setlogmask( int );
+static void prepend(char *buf, int bufsize, char *tag);
+
 static void 
 closelog_intern(int to_default)
 {
@@ -142,16 +163,15 @@ syslog(int pri, const char *fmt, ...)
 }
 
 void
-vsyslog(pri, fmt, ap)
-	int pri;
-	const char *fmt;
-	va_list ap;
+vsyslog( int pri, const char *fmt, va_list ap )
 {
 	register int cnt;
 	register char *p;
+	char *last_chr;
 	time_t now;
 	int fd, saved_errno;
-	char tbuf[2048], fmt_cpy[1024],*stdp;
+	int sz, rc;
+	char tbuf[2048], fmt_cpy[1024], *stdp;
 
 	saved_errno = errno;
 
@@ -173,6 +193,7 @@ vsyslog(pri, fmt, ap)
 	/* Build the message. */
 	(void)time(&now);
 	(void)snprintf(tbuf, sizeof(tbuf), "<%d>%.15s ", pri, ctime(&now) + 4);
+	stdp = "";
 	for (p = tbuf; *p; ++p);
 	if (LogStat & LOG_PERROR)
 		stdp = p;
@@ -187,26 +208,57 @@ vsyslog(pri, fmt, ap)
 	if (LogTag) {
 		*p++ = ':';
 		*p++ = ' ';
+		*p = '\0';
 	}
 
 	/* Substitute error message for %m. */
 	{
 		register char ch, *t1;
-		char *strerror();
 
 		/* We have to make sure we don't overrun fmt_cpy. */
-		for (t1 = fmt_cpy; (ch = *fmt) != '\0' && t1<fmt_cpy+sizeof(fmt_cpy); ++fmt)
+		for (t1 = fmt_cpy; ((ch=*fmt)!='\0') && (t1<(fmt_cpy+(sizeof(fmt_cpy)-1))); ++fmt) {
+			if (ch == '%' && fmt[1] == '%') {
+				if (t1<(fmt_cpy+(sizeof(fmt_cpy)-2))) {
+					*t1++='%'; *t1++='%'; fmt++;
+					continue;
+				} else {
+					/* flag message as truncated to alert the good guys */
+					prepend(tbuf, sizeof(tbuf), TRUNCATED); 
+					for (; *p; ++p);
+					break;
+				}
+			} 
 			if (ch == '%' && fmt[1] == 'm') {
 				++fmt;
-				t1 += snprintf(t1, sizeof(fmt_cpy)-(t1-fmt_cpy),
-					       "%s", strerror(saved_errno));
+				sz=sizeof(fmt_cpy)-(t1-fmt_cpy);
+				rc=snprintf(t1, sz, "%s", strerror(saved_errno));
+				if ((-1==rc) || (rc>sz)) {
+					/* flag message as truncated to alert the good guys */
+					prepend(tbuf, sizeof(tbuf), TRUNCATED); 
+					for (; *p; ++p);
+					for (; *t1; t1++);
+					break;
+				}
+				for (; *t1; t1++);
 			}
 			else
 				*t1++ = ch;
+		}
 		*t1 = '\0';
+		if (ch!=0) {
+			/* flag message as truncated to alert the good guys */
+			prepend(tbuf, sizeof(tbuf), TRUNCATED); 
+			for (; *p; ++p);
+		}
 	}
 
-	p += vsnprintf(p, sizeof(tbuf)-(p-tbuf), fmt_cpy, ap);
+	sz=sizeof(tbuf)-(p-tbuf);
+	rc=vsnprintf(p, sz, fmt_cpy, ap);
+	if ((-1==rc) || (rc>sz)) {
+		/* flag message as truncated to alert the good guys */
+		prepend(tbuf, sizeof(tbuf), TRUNCATED); 
+	}
+	for(; *p; p++);
 	cnt = p - tbuf;
 
 	/* Output to stderr if requested. */
@@ -225,17 +277,24 @@ vsyslog(pri, fmt, ap)
 	/* Output the message to the local logger. */
 #if 1
 	/* Use NUL as a message delimiter. */
-	if (write(LogFile, tbuf, cnt + 1) >= 1)
+	p=tbuf; last_chr=tbuf+cnt;
+	do {
+		rc=write(LogFile, p, ((last_chr-p)+1));
+		if (rc<0) {
+			if ((errno==EAGAIN) || (errno==EINTR))
+				rc=0;
+			else {
+				closelog_intern(0);
+				break;
+			}
+		}
+		p+=rc;
+	} while (p<=last_chr);
+	if (p>last_chr) 
 	{
-	        UNLOCK();
+		UNLOCK();
 		return;
 	}
-	else
-	{
-		/* If the write fails, we try to reconnect it next
-		 * time. */
-		closelog_intern (0);
-	}
 #else
 	if (write(LogFile, tbuf, cnt) >= 0) {
                 UNLOCK();
@@ -265,9 +324,7 @@ static struct sockaddr SyslogAddr;	/* AF_UNIX address of local logger */
  * OPENLOG -- open system log
  */
 void
-openlog(ident, logstat, logfac)
-	const char *ident;
-	int logstat, logfac;
+openlog( const char *ident, int logstat, int logfac )
 {
         LOCK();
 
@@ -306,7 +363,7 @@ openlog(ident, logstat, logfac)
  * CLOSELOG -- close the system log
  */
 void
-closelog()
+closelog( void )
 {
 	closelog_intern(1);
 }
@@ -315,8 +372,7 @@ closelog()
  * SETLOGMASK -- set the log mask level
  */
 int
-setlogmask(pmask)
-	int pmask;
+setlogmask( int pmask )
 {
 	int omask;
 
@@ -327,3 +383,52 @@ setlogmask(pmask)
 	UNLOCK();
 	return (omask);
 }
+
+static void prepend(char *buf, int bufsize, char *tag)
+{
+
+	/* Prepends "tag" to the beginning of the data in "buf" 
+	   right after the priority (if present) and the timestamp
+	   (if present) shifting existing contents to the right 
+	   to make room, and truncating any characters which may 
+	   get shifted off the end of "buf". If the tag is already
+	   there, we return without doing anything.
+	   A. Ian Vogelesang <vogelesang@hdshq.com> March 7/96 */
+
+	/* bufsize is total size of buf, including terminating null */
+	/* tag must be null-terminated string */
+
+	int taglen,bufleft;
+	char *p;
+	register char *q;
+
+	if ((NULL==buf) || (NULL==tag)) return;
+	taglen=strlen(tag);
+	if (0==taglen) return;
+
+	p=buf;
+
+	if ((p[0]=='<') && isdigit(p[1])) {
+		if (p[2]=='>') p+=3;
+		else if (isdigit(p[2]))
+			if (p[3]=='>') p+=4;
+			else if (isdigit(p[3]) && (p[4]=='>')) p+=5;
+	}
+	if (isalpha(p[0]) && isalpha(p[1]) && isalpha(p[2]) && \
+	    (' '==p[3]) && ((' '==p[4]) || isdigit(p[4])) && isdigit(p[5]) && \
+	    (' '==p[6]) && isdigit(p[7]) && isdigit(p[8]) && \
+	    (':'==p[9]) && isdigit(p[10]) && isdigit(p[11]) && \
+	    (':'==p[12]) && isdigit(p[13]) && isdigit(p[14]) && \
+	    (' '==p[15]) ) p+=16;
+	bufleft=bufsize-(p-buf);
+	if (taglen>=(bufleft-1)) {
+		bcopy(tag,p,(bufleft-1));
+		*(p+(bufleft-1))='\0';
+		return;
+	}
+	if (0==bcmp(p,tag,taglen)) return;
+	*(p+(bufleft-1)-taglen)='\0';
+	for (q=index(p,'\0'); q>=p; q--) *(q+taglen)=*q;
+	bcopy(tag,p,taglen);
+	return;
+}
