@@ -1,5 +1,5 @@
 /* Read decimal floating point numbers.
-Copyright (C) 1995 Free Software Foundation, Inc.
+Copyright (C) 1995, 1996 Free Software Foundation, Inc.
 Contributed by Ulrich Drepper.
 
 This file is part of the GNU C Library.
@@ -106,7 +106,7 @@ static const mp_limb _tens_in_limb[MAX_DIG_PER_LIMB + 1] =
 #define	MPNSIZE		(howmany (MAX_EXP + 2 * MANT_DIG, BITS_PER_MP_LIMB) \
 			 + 2)
 /* Declare an mpn integer variable that big.  */
-#define	MPN_VAR(name)	mp_limb name[MPNSIZE]; mp_size_t name##size
+#define	MPN_VAR(name)	mp_limb name[MPNSIZE] = {0}; mp_size_t name##size
 /* Copy an mpn integer value.  */
 #define MPN_ASSIGN(dst, src) \
 	memcpy (dst, src, (dst##size = src##size) * sizeof (mp_limb))
@@ -539,12 +539,12 @@ INTERNAL (STRTOF) (nptr, endptr, group)
 	      c = *++cp;
 	    }
 	  while (isdigit (c));
+
+	  if (exp_negative)
+	    exponent = -exponent;
 	}
       else
 	cp = expp;
-
-      if (exp_negative)
-	exponent = -exponent;
     }
 
   /* We don't want to have to work with trailing zeroes after the radix.  */
@@ -1126,9 +1126,13 @@ __STRTOF
 #else
 STRTOF
 #endif
+#ifdef __STDC__
+   ( const char *nptr, char **endptr)
+#else
 	(nptr, endptr)
      const char *nptr;
      char **endptr;
+#endif
 {
   return INTERNAL (STRTOF) (nptr, endptr, 0);
 }
