@@ -21,5 +21,9 @@ accept(int sockfd, const struct sockaddr *peer, int *paddrlen)
 }
 
 #ifdef __ELF__
+#ifdef __GNUC__
+int accept ( int, const struct sockaddr *, int * ) __attribute__  ((weak));
+#else
 #pragma weak accept
-#endif
+#endif   /* __GNUC__ */
+#endif   /* __ELF__ */
