@@ -23,5 +23,9 @@ send (int sockfd, const void *buffer, size_t len, unsigned flags)
 }
 
 #ifdef __ELF__
+#ifdef __GNUC__
+int send (int, const void *, size_t, unsigned) __attribute__ ((weak));
+#else
 #pragma weak send
-#endif
+#endif   /* __GNUC__ */
+#endif   /* __ELF__ */
